@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 1996-2016. All Rights Reserved.
+%% Copyright Ericsson AB 1996-2018. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -81,7 +81,7 @@
 
 -record(rule, {
           n,             % rule n in the grammar file
-          line,
+          anno,
           symbols,       % the names of symbols
           tokens,
           is_guard,      % the action is a guard (not used)
@@ -105,7 +105,7 @@
 
 -record(user_code, {state, terminal, funname, action}).
 
--record(symbol, {line = none, name}).
+-record(symbol, {anno = none, name}).
 
 %% ACCEPT is neither an atom nor a non-terminal.
 -define(ACCEPT, {}).
@@ -154,13 +154,13 @@ compile(Input0, Output0,
 format_error(bad_declaration) ->
     io_lib:fwrite("unknown or bad declaration, ignored", []);
 format_error({bad_expect, SymName}) ->
-    io_lib:fwrite("argument ~s of Expect is not an integer", 
+    io_lib:fwrite("argument ~ts of Expect is not an integer", 
                   [format_symbol(SymName)]);
 format_error({bad_rootsymbol, SymName}) ->
-    io_lib:fwrite("rootsymbol ~s is not a nonterminal", 
+    io_lib:fwrite("rootsymbol ~ts is not a nonterminal", 
                   [format_symbol(SymName)]);
 format_error({bad_states, SymName}) ->
-    io_lib:fwrite("argument ~s of States is not an integer", 
+    io_lib:fwrite("argument ~ts of States is not an integer", 
                   [format_symbol(SymName)]);
 format_error({conflict, Conflict}) ->
     format_conflict(Conflict);
@@ -169,19 +169,19 @@ format_error({conflicts, SR, RR}) ->
 format_error({duplicate_declaration, Tag}) ->
     io_lib:fwrite("duplicate declaration of ~s", [atom_to_list(Tag)]);
 format_error({duplicate_nonterminal, Nonterminal}) ->
-    io_lib:fwrite("duplicate non-terminals ~s", 
+    io_lib:fwrite("duplicate non-terminals ~ts", 
                   [format_symbol(Nonterminal)]);
 format_error({duplicate_precedence, Op}) ->
-    io_lib:fwrite("duplicate precedence operator ~s", 
+    io_lib:fwrite("duplicate precedence operator ~ts", 
                   [format_symbol(Op)]);
 format_error({duplicate_terminal, Terminal}) ->
-    io_lib:fwrite("duplicate terminal ~s", 
+    io_lib:fwrite("duplicate terminal ~ts", 
                   [format_symbol(Terminal)]);
 format_error({endsymbol_is_nonterminal, Symbol}) ->
-    io_lib:fwrite("endsymbol ~s is a nonterminal", 
+    io_lib:fwrite("endsymbol ~ts is a nonterminal", 
                   [format_symbol(Symbol)]);
 format_error({endsymbol_is_terminal, Symbol}) ->
-    io_lib:fwrite("endsymbol ~s is a terminal", 
+    io_lib:fwrite("endsymbol ~ts is a terminal", 
                   [format_symbol(Symbol)]);
 format_error({error, Module, Error}) ->
     Module:format_error(Error);
@@ -192,7 +192,7 @@ format_error(illegal_empty) ->
 format_error({internal_error, Error}) ->
     io_lib:fwrite("internal yecc error: ~w", [Error]);
 format_error({missing_syntax_rule, Nonterminal}) ->
-    io_lib:fwrite("no syntax rule for non-terminal symbol ~s",
+    io_lib:fwrite("no syntax rule for non-terminal symbol ~ts",
                   [format_symbol(Nonterminal)]);
 format_error({n_states, Exp, N}) ->
     io_lib:fwrite("expected ~w states, but got ~p states", [Exp, N]);
@@ -201,31 +201,31 @@ format_error(no_grammar_rules) ->
 format_error(nonterminals_missing) ->
     io_lib:fwrite("Nonterminals is missing", []);
 format_error({precedence_op_is_endsymbol, SymName}) ->
-    io_lib:fwrite("precedence operator ~s is endsymbol",
+    io_lib:fwrite("precedence operator ~ts is endsymbol",
                   [format_symbol(SymName)]);
 format_error({precedence_op_is_unknown, SymName}) ->
-    io_lib:fwrite("unknown precedence operator ~s",
+    io_lib:fwrite("unknown precedence operator ~ts",
                   [format_symbol(SymName)]);
 format_error({reserved, N}) ->
     io_lib:fwrite("the use of ~w should be avoided", [N]);
 format_error({symbol_terminal_and_nonterminal, SymName}) ->
-    io_lib:fwrite("symbol ~s is both a terminal and nonterminal",
+    io_lib:fwrite("symbol ~ts is both a terminal and nonterminal",
                   [format_symbol(SymName)]);
 format_error(rootsymbol_missing) ->
     io_lib:fwrite("Rootsymbol is missing", []);
 format_error(terminals_missing) ->
     io_lib:fwrite("Terminals is missing", []);
 format_error({undefined_nonterminal, Symbol}) ->
-    io_lib:fwrite("undefined nonterminal: ~s", [format_symbol(Symbol)]);
+    io_lib:fwrite("undefined nonterminal: ~ts", [format_symbol(Symbol)]);
 format_error({undefined_pseudo_variable, Atom}) ->
     io_lib:fwrite("undefined pseudo variable ~w", [Atom]);
 format_error({undefined_symbol, SymName}) ->
-    io_lib:fwrite("undefined rhs symbol ~s", [format_symbol(SymName)]);
+    io_lib:fwrite("undefined rhs symbol ~ts", [format_symbol(SymName)]);
 format_error({unused_nonterminal, Nonterminal}) ->
-    io_lib:fwrite("non-terminal symbol ~s not used", 
+    io_lib:fwrite("non-terminal symbol ~ts not used", 
                   [format_symbol(Nonterminal)]);
 format_error({unused_terminal, Terminal}) ->
-    io_lib:fwrite("terminal symbol ~s not used", 
+    io_lib:fwrite("terminal symbol ~ts not used", 
                   [format_symbol(Terminal)]);
 format_error({bad_symbol, String}) ->
     io_lib:fwrite("bad symbol ~ts", [String]);
@@ -365,10 +365,10 @@ is_filename(T) ->
 
 shorten_filename(Name0) ->
     {ok,Cwd} = file:get_cwd(),
-    case lists:prefix(Cwd, Name0) of
-        false -> Name0;
-        true ->
-            case lists:nthtail(length(Cwd), Name0) of
+    case string:prefix(Name0, Cwd) of
+        nomatch -> Name0;
+        Rest ->
+            case unicode:characters_to_list(Rest) of
                 "/"++N -> N;
                 N -> N
             end
@@ -455,10 +455,14 @@ os_process_size() ->
     case os:type() of
         {unix, sunos} ->
             Size = os:cmd("ps -o vsz -p " ++ os:getpid() ++ " | tail -1"),
-            list_to_integer(lib:nonl(Size));
+            list_to_integer(nonl(Size));
         _ ->
             0
-    end.            
+    end.
+
+nonl([$\n]) -> [];
+nonl([]) -> [];
+nonl([H|T]) -> [H|nonl(T)].
 
 timeit(Name, Fun, St0) ->
     Time = runtime,
@@ -517,7 +521,7 @@ parse_grammar(Grammar, Inport, NextLine, St0) ->
     parse_grammar(Inport, NextLine, St).
 
 parse_grammar({error,ErrorLine,Error}, St) ->
-    add_error(ErrorLine, Error, St);
+    add_error(erl_anno:new(ErrorLine), Error, St);
 parse_grammar({rule, Rule, Tokens}, St0) ->
     NmbrOfDaughters = case Rule of
                           [_, #symbol{name = '$empty'}]  -> 0;
@@ -534,15 +538,15 @@ parse_grammar({rule, Rule, Tokens}, St0) ->
     St#yecc{rules_list = [RuleDef | St#yecc.rules_list]};
 parse_grammar({prec, Prec}, St) ->
     St#yecc{prec = Prec ++ St#yecc.prec};
-parse_grammar({#symbol{}, [{string,Line,String}]}, St) ->
-    add_error(Line, {bad_symbol, String}, St);
-parse_grammar({#symbol{line = Line, name = Name}, Symbols}, St) ->
+parse_grammar({#symbol{}, [{string,Anno,String}]}, St) ->
+    add_error(Anno, {bad_symbol, String}, St);
+parse_grammar({#symbol{anno = Anno, name = Name}, Symbols}, St) ->
     CF = fun(I) ->
                  case element(I, St) of
                      [] -> 
                          setelement(I, St, Symbols);
                      _ -> 
-                         add_error(Line, {duplicate_declaration, Name}, St)
+                         add_error(Anno, {duplicate_declaration, Name}, St)
                  end
          end,
     OneSymbol = length(Symbols) =:= 1,
@@ -553,7 +557,7 @@ parse_grammar({#symbol{line = Line, name = Name}, Symbols}, St) ->
         'Endsymbol' when OneSymbol ->  CF(#yecc.endsymbol);
         'Expect' when OneSymbol -> CF(#yecc.expect_shift_reduce);
         'States' when OneSymbol -> CF(#yecc.expect_n_states); % undocumented
-        _ -> add_warning(Line, bad_declaration, St)
+        _ -> add_warning(Anno, bad_declaration, St)
     end.
 
 read_grammar(Inport, St, Line) ->
@@ -599,7 +603,7 @@ precedence(_) -> unknown.
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
 check_grammar(St0) ->
-    Empty = #symbol{line = none, name = '$empty'},
+    Empty = #symbol{anno = none, name = '$empty'},
     AllSymbols = St0#yecc.nonterminals ++ St0#yecc.terminals ++ [Empty],
     St1 = St0#yecc{all_symbols = AllSymbols},
     Cs = [fun check_nonterminals/1, fun check_terminals/1, 
@@ -640,12 +644,12 @@ check_rootsymbol(St) ->
     case St#yecc.rootsymbol of
         [] ->
             add_error(rootsymbol_missing, St);
-        [#symbol{line = Line, name = SymName}] ->
+        [#symbol{anno = Anno, name = SymName}] ->
             case kind_of_symbol(St, SymName) of
                 nonterminal ->
                     St#yecc{rootsymbol = SymName};
                 _ ->
-                    add_error(Line, {bad_rootsymbol, SymName}, St)
+                    add_error(Anno, {bad_rootsymbol, SymName}, St)
             end
     end.
 
@@ -653,12 +657,12 @@ check_endsymbol(St) ->
     case St#yecc.endsymbol of
         [] ->
             St#yecc{endsymbol = '$end'};
-        [#symbol{line = Line, name = SymName}] ->
+        [#symbol{anno = Anno, name = SymName}] ->
             case kind_of_symbol(St, SymName) of
                 nonterminal ->
-                    add_error(Line, {endsymbol_is_nonterminal, SymName}, St);
+                    add_error(Anno, {endsymbol_is_nonterminal, SymName}, St);
                 terminal ->
-                    add_error(Line, {endsymbol_is_terminal, SymName}, St);
+                    add_error(Anno, {endsymbol_is_terminal, SymName}, St);
                 _ ->
                     St#yecc{endsymbol = SymName}
             end
@@ -670,8 +674,8 @@ check_expect(St0) ->
             St0#yecc{expect_shift_reduce = 0};
         [#symbol{name = Expect}] when is_integer(Expect) ->
             St0#yecc{expect_shift_reduce = Expect};
-        [#symbol{line = Line, name = Name}] ->
-            St1 = add_error(Line, {bad_expect, Name}, St0),
+        [#symbol{anno = Anno, name = Name}] ->
+            St1 = add_error(Anno, {bad_expect, Name}, St0),
             St1#yecc{expect_shift_reduce = 0}
     end.
 
@@ -681,27 +685,27 @@ check_states(St) ->
             St;
         [#symbol{name = NStates}] when is_integer(NStates) ->
             St#yecc{expect_n_states = NStates};
-        [#symbol{line = Line, name = Name}] ->
-            add_error(Line, {bad_states, Name}, St)
+        [#symbol{anno = Anno, name = Name}] ->
+            add_error(Anno, {bad_states, Name}, St)
     end.
 
 check_precedences(St0) ->
     {St1, _} = 
-        foldr(fun({#symbol{line = Line, name = Op},_I,_A}, {St,Ps}) ->
+        foldr(fun({#symbol{anno = Anno, name = Op},_I,_A}, {St,Ps}) ->
                       case member(Op, Ps) of
                           true ->
-                              {add_error(Line, {duplicate_precedence,Op}, St),
+                              {add_error(Anno, {duplicate_precedence,Op}, St),
                                Ps};
                           false ->
                               {St, [Op | Ps]}
                       end
               end, {St0,[]}, St0#yecc.prec),
-    foldl(fun({#symbol{line = Line, name = Op},I,A}, St) ->
+    foldl(fun({#symbol{anno = Anno, name = Op},I,A}, St) ->
                   case kind_of_symbol(St, Op) of
                       endsymbol ->
-                          add_error(Line,{precedence_op_is_endsymbol,Op}, St);
+                          add_error(Anno,{precedence_op_is_endsymbol,Op}, St);
                       unknown ->
-                          add_error(Line, {precedence_op_is_unknown, Op}, St);
+                          add_error(Anno, {precedence_op_is_unknown, Op}, St);
                       _ -> 
                           St#yecc{prec = [{Op,I,A} | St#yecc.prec]}
                   end
@@ -709,13 +713,13 @@ check_precedences(St0) ->
 
 check_rule(Rule0, {St0,Rules}) ->
     Symbols = Rule0#rule.symbols,
-    #symbol{line = HeadLine, name = Head} = hd(Symbols),
+    #symbol{anno = HeadAnno, name = Head} = hd(Symbols),
     case member(Head, St0#yecc.nonterminals) of
         false -> 
-            {add_error(HeadLine, {undefined_nonterminal, Head}, St0), Rules};
+            {add_error(HeadAnno, {undefined_nonterminal, Head}, St0), Rules};
         true ->
             St = check_rhs(tl(Symbols), St0),
-            Rule = Rule0#rule{line = HeadLine, symbols = names(Symbols)},
+            Rule = Rule0#rule{anno = HeadAnno, symbols = names(Symbols)},
             {St, [Rule | Rules]}
     end.
 
@@ -725,7 +729,7 @@ check_rules(St0) ->
         [] ->
             add_error(no_grammar_rules, St);
         _ ->
-            Rule = #rule{line = none, 
+            Rule = #rule{anno = none,
                          symbols = [?ACCEPT, St#yecc.rootsymbol],
                          tokens = []},
             Rules1 = [Rule | Rules0],
@@ -740,9 +744,9 @@ duplicates(List) ->
 names(Symbols) ->
     map(fun(Symbol) -> Symbol#symbol.name end, Symbols).
 
-symbol_line(Name, St) ->
-    #symbol{line = Line} = symbol_find(Name, St#yecc.all_symbols),
-    Line.
+symbol_anno(Name, St) ->
+    #symbol{anno = Anno} = symbol_find(Name, St#yecc.all_symbols),
+    Anno.
 
 symbol_member(Symbol, Symbols) ->
     symbol_find(Symbol#symbol.name, Symbols) =/= false.
@@ -894,31 +898,33 @@ report_warnings(St) ->
 add_error(E, St) ->
     add_error(none, E, St).
 
-add_error(Line, E, St) ->
-    add_error(St#yecc.infile, Line, E, St).
+add_error(Anno, E, St) ->
+    add_error(St#yecc.infile, Anno, E, St).
 
-add_error(File, Line, E, St) ->
-    St#yecc{errors = [{File,{Line,?MODULE,E}}|St#yecc.errors]}.    
+add_error(File, Anno, E, St) ->
+    Loc = location(Anno),
+    St#yecc{errors = [{File,{Loc,?MODULE,E}}|St#yecc.errors]}.
 
 add_errors(SymNames, E0, St0) ->
     foldl(fun(SymName, St) ->
-                  add_error(symbol_line(SymName, St), {E0, SymName}, St)
+                  add_error(symbol_anno(SymName, St), {E0, SymName}, St)
           end, St0, SymNames).
 
-add_warning(Line, W, St) ->
-    St#yecc{warnings = [{St#yecc.infile,{Line,?MODULE,W}}|St#yecc.warnings]}.
+add_warning(Anno, W, St) ->
+    Loc = location(Anno),
+    St#yecc{warnings = [{St#yecc.infile,{Loc,?MODULE,W}}|St#yecc.warnings]}.
 
 add_warnings(SymNames, W0, St0) ->
     foldl(fun(SymName, St) ->
-                  add_warning(symbol_line(SymName, St), {W0, SymName}, St)
+                  add_warning(symbol_anno(SymName, St), {W0, SymName}, St)
           end, St0, SymNames).
 
 check_rhs([#symbol{name = '$empty'}], St) ->
     St;
 check_rhs(Rhs, St0) ->
     case symbol_find('$empty', Rhs) of
-        #symbol{line = Line} ->
-            add_error(Line, illegal_empty, St0);
+        #symbol{anno = Anno} ->
+            add_error(Anno, illegal_empty, St0);
         false ->
             foldl(fun(Sym, St) ->
                           case symbol_member(Sym, St#yecc.all_symbols) of
@@ -926,13 +932,13 @@ check_rhs(Rhs, St0) ->
                                   St;
                               false -> 
                                   E = {undefined_symbol,Sym#symbol.name},
-                                  add_error(Sym#symbol.line, E, St)
+                                  add_error(Sym#symbol.anno, E, St)
                           end
                   end, St0, Rhs)
     end.
 
 check_action(Tokens) ->
-    case erl_parse:parse_exprs(add_roberts_dot(Tokens, 0)) of
+    case erl_parse:parse_exprs(add_roberts_dot(Tokens, erl_anno:new(0))) of
         {error, _Error} ->
             {false, false};
         {ok, [Expr | Exprs]} ->
@@ -940,10 +946,10 @@ check_action(Tokens) ->
             {IsGuard, true}
     end.
 
-add_roberts_dot([], Line) ->
-    [{'dot', Line}];
-add_roberts_dot([{'dot', Line} | _], _) ->
-    [{'dot', Line}];
+add_roberts_dot([], Anno) ->
+    [{'dot', Anno}];
+add_roberts_dot([{'dot', Anno} | _], _) ->
+    [{'dot', Anno}];
 add_roberts_dot([Token | Tokens], _) ->
     [Token | add_roberts_dot(Tokens, element(2, Token))].
 
@@ -953,21 +959,22 @@ subst_pseudo_vars([H0 | T0], NmbrOfDaughters, St0) ->
     {H, St1} = subst_pseudo_vars(H0, NmbrOfDaughters, St0),
     {T, St} = subst_pseudo_vars(T0, NmbrOfDaughters, St1),
     {[H | T], St};
-subst_pseudo_vars({atom, Line, Atom}, NmbrOfDaughters, St0) ->
+subst_pseudo_vars({atom, Anno, Atom}, NmbrOfDaughters, St0) ->
     case atom_to_list(Atom) of
         [$$ | Rest] ->
             try list_to_integer(Rest) of
                 N when N > 0, N =< NmbrOfDaughters ->
-                    {{var, Line, list_to_atom(append("__", Rest))}, St0};
+                    {{var, Anno, list_to_atom(append("__", Rest))}, St0};
                 _ ->
-                    St = add_error(Line, {undefined_pseudo_variable, Atom}, 
+                    St = add_error(Anno,
+                                   {undefined_pseudo_variable, Atom},
                                    St0),
-                    {{atom, Line, '$undefined'}, St}
+                    {{atom, Anno, '$undefined'}, St}
             catch 
-                error: _ -> {{atom, Line, Atom}, St0}
+                error: _ -> {{atom, Anno, Atom}, St0}
             end;
         _ ->
-            {{atom, Line, Atom}, St0}
+            {{atom, Anno, Atom}, St0}
     end;
 subst_pseudo_vars(Tuple, NmbrOfDaughters, St0) when is_tuple(Tuple) ->
     {L, St} = subst_pseudo_vars(tuple_to_list(Tuple), NmbrOfDaughters, St0),
@@ -1338,7 +1345,7 @@ make_rule_pointer_info(StC, RpRhs, RuleIndex) ->
 rp_info([], _SymbolTab, _LcTab, _RuleIndex) ->
     [];
 rp_info([Category | Followers], SymbolTab, LcTab, RuleIndex) ->
-    case dict:find(Category, RuleIndex) of
+    case maps:find(Category, RuleIndex) of
         error -> % terminal
             [];
         {ok, ExpandingRules} when Followers =:= [] ->
@@ -1357,7 +1364,7 @@ rp_info([Category | Followers], SymbolTab, LcTab, RuleIndex) ->
 make_lookahead([], _, _, LA) ->
     {empty, LA};
 make_lookahead([Symbol | Symbols], SymbolTab, LcTab, LA) ->
-    case dict:find(Symbol, LcTab) of
+    case maps:find(Symbol, LcTab) of
         {ok, LeftCorner} -> % nonterminal
             case empty_member(LeftCorner) of
                 true ->
@@ -1370,7 +1377,7 @@ make_lookahead([Symbol | Symbols], SymbolTab, LcTab, LA) ->
             set_add(Symbol, LA)
     end.
 
-%% -> dict-of({Nonterminal, [Terminal]}).
+%% -> map-of({Nonterminal, [Terminal]}).
 %% The algorithm FIRST/1 from the Dragon Book.
 %% Left corner table, all terminals (including '$empty') that can
 %% begin strings generated by Nonterminal.
@@ -1379,15 +1386,15 @@ make_left_corner_table(#yecc{rules_list = RulesList} = St) ->
     Rules = map(fun(#rule{symbols = [Lhs | Rhs]}) ->
                         {Lhs,{Lhs, Rhs}}
                 end, RulesList),
-    LeftHandTab = dict:from_list(family(Rules)),
+    LeftHandTab = maps:from_list(family(Rules)),
     X0 = [{S,H} || {H,{H,Rhs}} <- Rules, 
                    S <- Rhs, 
                    not is_terminal(SymbolTab, S)],
     XL = family_with_domain(X0, St#yecc.nonterminals),
-    X = dict:from_list(XL),
-    Xref = fun(NT) -> dict:fetch(NT, X) end,
+    X = maps:from_list(XL),
+    Xref = fun(NT) -> maps:get(NT, X) end,
     E = set_empty(),
-    LC0 = dict:from_list([{H, E} || {H,_} <- XL]),
+    LC0 = maps:from_list([{H, E} || {H,_} <- XL]),
     %% Handle H -> a S, where a is a terminal ('$empty' inclusive).
     {Q, LC1} =
         foldl(fun({H,{H,[S | _]}}, {Q0, LC}) ->
@@ -1406,7 +1413,7 @@ left_corners(Q0, LC0, LeftHandTab, SymbolTab, Xref) ->
         [] -> 
             LC0;
         Q1 -> 
-            Rs = flatmap(fun(NT) -> dict:fetch(NT, LeftHandTab) end, Q1),
+            Rs = flatmap(fun(NT) -> maps:get(NT, LeftHandTab) end, Q1),
             {LC, Q} = left_corners2(Rs, LC0, [], SymbolTab, Xref),
             left_corners(Q, LC, LeftHandTab, SymbolTab, Xref)
     end.
@@ -1415,7 +1422,7 @@ left_corners2([], LC, Q, _SymbolTab, _Xref) ->
     {LC, Q};
 left_corners2([{Head,Rhs} | Rs], LC, Q0, SymbolTab, Xref) ->
     Ts = left_corner_rhs(Rhs, Head, LC, set_empty(), SymbolTab),
-    First0 = dict:fetch(Head, LC),
+    First0 = maps:get(Head, LC),
     case set_is_subset(Ts, First0) of
         true ->
             left_corners2(Rs, LC, Q0, SymbolTab, Xref);
@@ -1425,14 +1432,14 @@ left_corners2([{Head,Rhs} | Rs], LC, Q0, SymbolTab, Xref) ->
     end.
 
 upd_first(NT, Ts, LC) ->
-    dict:update(NT, fun(First) -> set_union(First, Ts) end, LC).
+    maps:update_with(NT, fun(First) -> set_union(First, Ts) end, LC).
 
 left_corner_rhs([S | Ss], Head, LC, Ts, SymbolTab) ->
     case ets:lookup(SymbolTab, S) of
         [{_,Num}=SymbolAndNum] when Num >= 0 ->
             set_add_terminal(SymbolAndNum, Ts);
         [_NonTerminalSymbol] ->
-            First = dict:fetch(S, LC),
+            First = maps:get(S, LC),
             case empty_member(First) of
                 true ->
                     NTs = set_union(empty_delete(First), Ts),
@@ -1459,7 +1466,7 @@ make_rule_index(#yecc{nonterminals = Nonterminals,
     Symbol2Rule = [{Foo,R} || #rule{symbols = Symbols}=R <- RulesListNoCodes,
                               Foo <- Symbols],
     Pointer2Rule = [{I, R} || {{_Foo,R},I} <- count(1, Symbol2Rule)],
-    {dict:from_list(IndexedTab), dict:from_list(Pointer2Rule)}.
+    {maps:from_list(IndexedTab), maps:from_list(Pointer2Rule)}.
 
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 % Computing parse action table from list of states and goto table:
@@ -1806,9 +1813,9 @@ report_conflict(Conflict, St, ActionName, How) ->
             Formated = format_symbol(ActionName),
             case How of 
                 prec ->
-                    io:fwrite(<<"Resolved in favor of ~s.\n\n">>, [Formated]);
+                    io:fwrite(<<"Resolved in favor of ~ts.\n\n">>, [Formated]);
                 default ->
-                    io:fwrite(<<"Conflict resolved in favor of ~s.\n\n">>, 
+                    io:fwrite(<<"Conflict resolved in favor of ~ts.\n\n">>, 
                               [Formated])
             end;
         true ->
@@ -1853,7 +1860,7 @@ format_conflict({Symbol, N, _, {one_level_up,
                                 {L1, RuleN1, {P1, Ass1}}, 
                                 {L2, RuleN2, {P2, Ass2}}}}) ->
     S1 = io_lib:fwrite(<<"Conflicting precedences of symbols when "
-                         "scanning ~s in state ~w:\n">>, 
+                         "scanning ~ts in state ~w:\n">>, 
                        [format_symbol(Symbol), N]),
     S2 = io_lib:fwrite(<<"   ~s ~w (rule ~w at line ~w)\n"
                           "      vs.\n">>,
@@ -1863,26 +1870,26 @@ format_conflict({Symbol, N, _, {one_level_up,
     [S1, S2, S3];
 format_conflict({Symbol, N, Reduce, Confl}) ->
     S1 = io_lib:fwrite(<<"Parse action conflict scanning symbol "
-                         "~s in state ~w:\n">>, [format_symbol(Symbol), N]),
+                         "~ts in state ~w:\n">>, [format_symbol(Symbol), N]),
     S2 = case Reduce of
              {[HR | TR], RuleNmbr, RuleLine} ->
-                 io_lib:fwrite(<<"   Reduce to ~s from ~s (rule ~w at "
+                 io_lib:fwrite(<<"   Reduce to ~ts from ~ts (rule ~w at "
                                  "line ~w)\n      vs.\n">>,
                                [format_symbol(HR), format_symbols(TR), 
                                 RuleNmbr, RuleLine])
          end,
     S3 = case Confl of 
              {reduce, [HR2|TR2], RuleNmbr2, RuleLine2} ->
-                 io_lib:fwrite(<<"   reduce to ~s from ~s "
+                 io_lib:fwrite(<<"   reduce to ~ts from ~ts "
                                  "(rule ~w at line ~w).">>,
                                [format_symbol(HR2), format_symbols(TR2), 
                                 RuleNmbr2, RuleLine2]);
              {shift, NewState, Sym} ->
                  io_lib:fwrite(<<"   shift to state ~w, adding right "
-                                 "sisters to ~s.">>,
+                                 "sisters to ~ts.">>,
                                [NewState, format_symbol(Sym)]);
              {accept, Rootsymbol} ->
-                 io_lib:fwrite(<<"   reduce to rootsymbol ~s.">>,
+                 io_lib:fwrite(<<"   reduce to rootsymbol ~ts.">>,
                                [format_symbol(Rootsymbol)])
          end,
     [S1, S2, S3].
@@ -1923,8 +1930,9 @@ format_conflict({Symbol, N, Reduce, Confl}) ->
 
 -define(CODE_VERSION, "1.4").
 -define(YECC_BUG(M, A), 
-        iolist_to_binary([" erlang:error({yecc_bug,\"",?CODE_VERSION,"\",",
-                          io_lib:fwrite(M, A), "}).\n\n"])).
+        unicode:characters_to_binary(
+          [" erlang:error({yecc_bug,\"",?CODE_VERSION,"\",",
+           io_lib:fwrite(M, A), "}).\n\n"])).
 
 %% Returns number of newlines in included files.
 output_prelude(Outport, Inport, St0) when St0#yecc.includefile =:= [] ->
@@ -1977,7 +1985,7 @@ output_header(St0) ->
 output_goto(St, [{_Nonterminal, []} | Go], StateInfo) ->
     output_goto(St, Go, StateInfo);
 output_goto(St0, [{Nonterminal, List} | Go], StateInfo) ->
-    F = function_name(yeccgoto, Nonterminal),
+    F = function_name(St0, yeccgoto, Nonterminal),
     St05 = fwrite(St0, <<"-dialyzer({nowarn_function, ~w/7}).\n">>, [F]),
     St10 = output_goto1(St05, List, F, StateInfo, true),
     St = output_goto_fini(F, Nonterminal, St10),
@@ -2015,7 +2023,8 @@ output_goto_fini(F, NT, #yecc{includefile_version = {1,1}}=St0) ->
     St = fwrite(St10, <<"~w(State, _Cat, _Ss, _Stack, _T, _Ts, _Tzr) ->\n">>,
                 [F]),
     fwrite(St, 
-           ?YECC_BUG(<<"{~w, State, missing_in_goto_table}">>, [NT]),
+           ?YECC_BUG(<<"{~ts, State, missing_in_goto_table}">>,
+                     [quoted_atom(St0, NT)]),
            []);
 output_goto_fini(_F, _NT, St) ->
     fwrite(St, <<".\n\n">>, []).
@@ -2024,7 +2033,7 @@ output_goto_fini(_F, _NT, St) ->
 find_user_code(ParseActions, St) ->
     [#user_code{state = State, 
                 terminal = Terminal, 
-                funname = inlined_function_name(State, Terminal), 
+                funname = inlined_function_name(St, State, Terminal),
                 action = Action} || 
         {State, La_actions} <- ParseActions,
         {Action, Terminals, RuleNmbr, NmbrOfDaughters} 
@@ -2145,14 +2154,14 @@ output_action(St, State, Terminal, #reduce{}=Action, IsFirst, SI) ->
     output_reduce(St, State, Terminal, Action, IsFirst, SI);
 output_action(St0, State, Terminal, #shift{state = NewState}, IsFirst, _SI) ->
     St10 = delim(St0, IsFirst),
-    St = fwrite(St10, <<"yeccpars2_~w(S, ~s, Ss, Stack, T, Ts, Tzr) ->\n">>,
-                [State, quoted_atom(Terminal)]),
+    St = fwrite(St10, <<"yeccpars2_~w(S, ~ts, Ss, Stack, T, Ts, Tzr) ->\n">>,
+                [State, quoted_atom(St10, Terminal)]),
     output_call_to_includefile(NewState, St);
 output_action(St0, State, Terminal, accept, IsFirst, _SI) ->
     St10 = delim(St0, IsFirst),
     St = fwrite(St10, 
-                <<"yeccpars2_~w(_S, ~s, _Ss, Stack, _T, _Ts, _Tzr) ->\n">>,
-                [State, quoted_atom(Terminal)]),
+                <<"yeccpars2_~w(_S, ~ts, _Ss, Stack, _T, _Ts, _Tzr) ->\n">>,
+                [State, quoted_atom(St10, Terminal)]),
     fwrite(St, <<" {ok, hd(Stack)}">>, []);
 output_action(St, _State, _Terminal, nonassoc, _IsFirst, _SI) ->
     St.
@@ -2171,19 +2180,19 @@ output_state_actions_fini(State, IsFirst, St0) ->
     St = fwrite(St10, <<"yeccpars2_~w(_, _, _, _, T, _, _) ->\n">>, [State]),
     fwrite(St, <<" yeccerror(T).\n\n">>, []).
 
-output_reduce(St0, State, Terminal0, 
+output_reduce(St0, State, Terminal,
               #reduce{rule_nmbr = RuleNmbr, 
                       head = Head, 
                       nmbr_of_daughters = NmbrOfDaughters},
               IsFirst, StateInfo) ->
     St10 = delim(St0, IsFirst),
-    Terminal = if 
-                   is_atom(Terminal0) -> quoted_atom(Terminal0);
-                   true -> Terminal0
-               end,
+    QuotedTerminal = if 
+                         is_atom(Terminal) -> quoted_atom(St10, Terminal);
+                         true -> Terminal
+                     end,
     St20 = fwrite(St10,
-                  <<"yeccpars2_~w(_S, ~s, Ss, Stack, T, Ts, Tzr) ->\n">>,
-                  [State, Terminal]),
+                  <<"yeccpars2_~w(_S, ~ts, Ss, Stack, T, Ts, Tzr) ->\n">>,
+                  [State, QuotedTerminal]),
     St30 = 
         if
             NmbrOfDaughters < 2 ->
@@ -2191,8 +2200,8 @@ output_reduce(St0, State, Terminal0,
                 St20;
             true ->
                 Ns = "Nss",
-                Tmp = string:join(lists:duplicate(NmbrOfDaughters - 1, "_"),
-                                  ","),
+                Tmp = lists:join(",",
+                                  lists:duplicate(NmbrOfDaughters - 1, "_")),
                 fwrite(St20, <<" [~s|Nss] = Ss,\n">>, [Tmp])
         end,
     St40 = case tokens(RuleNmbr, St30) of
@@ -2202,7 +2211,7 @@ output_reduce(St0, State, Terminal0,
                _ ->
                    NewStack = "NewStack",
                    fwrite(St30, <<" NewStack = ~w(Stack),\n">>, 
-                          [inlined_function_name(State, Terminal0)])
+                          [inlined_function_name(St30, State, Terminal)])
                end,
     if 
         NmbrOfDaughters =:= 0 ->
@@ -2218,13 +2227,13 @@ output_reduce(St0, State, Terminal0,
             St = fwrite(St40, <<"~s">>, [C]),
             %% Short-circuit call to yeccpars2:
             fwrite(St,
-                   <<" yeccpars2_~w(~s, ~s, [~w | Ss], ~s, T, Ts, Tzr)">>,
-                   [Repr, NextS, Terminal, State, NewStack]);
+                   <<" yeccpars2_~w(~s, ~ts, [~w | Ss], ~s, T, Ts, Tzr)">>,
+                   [Repr, NextS, QuotedTerminal, State, NewStack]);
         true ->
             fwrite(St40, 
-                   <<" ~w(hd(~s), ~s, ~s, ~s, T, Ts, Tzr)">>,
-                   [function_name(yeccgoto, Head), Ns,
-                    Terminal, Ns, NewStack])
+                   <<" ~w(hd(~s), ~ts, ~s, ~s, T, Ts, Tzr)">>,
+                   [function_name(St40, yeccgoto, Head), Ns,
+                    QuotedTerminal, Ns, NewStack])
     end.
 
 delim(St, true) ->
@@ -2232,8 +2241,10 @@ delim(St, true) ->
 delim(St, false) ->
     fwrite(St, <<";\n">>, []).
 
-quoted_atom(Atom) ->
-    io_lib:fwrite(<<"~w">>, [Atom]).
+quoted_atom(#yecc{encoding = latin1}, Atom) when is_atom(Atom) ->
+    io_lib:write_atom_as_latin1(Atom);
+quoted_atom(_St, Atomic) ->
+    io_lib:write(Atomic).
     
 output_inlined(St, UserCodeActions, Infile) ->
     foldl(fun(#user_code{funname = InlinedFunctionName, 
@@ -2285,25 +2296,27 @@ output_inlined(St0, FunctionName, Reduce, Infile) ->
     fwrite(St, <<" [begin\n  ~ts\n  end | ~s].\n\n">>,
            [pp_tokens(Tokens, Line0, St#yecc.encoding), Stack]).
 
-inlined_function_name(State, "Cat") ->
-    inlined_function_name(State, "");
-inlined_function_name(State, Terminal) ->
-    list_to_atom(concat([yeccpars2_, State, '_', Terminal])).
+inlined_function_name(St, State, Terminal) ->
+    End = case Terminal of
+              "Cat" -> [];
+              _ -> [quoted_atom(St, Terminal)]
+          end,
+    list_to_atom(concat([yeccpars2_, State, '_'] ++ End)).
 
--compile({nowarn_unused_function,function_name/2}).
-function_name(Name, Suf) ->
-    list_to_atom(concat([Name, '_' | quoted_atom(Suf)])).
+-compile({nowarn_unused_function,function_name/3}).
+function_name(St, Name, Suf) ->
+    list_to_atom(concat([Name, '_'] ++ [quoted_atom(St, Suf)])).
 
 rule(RulePointer, St) ->
-    #rule{n = N, line = Line, symbols = Symbols} = 
-        dict:fetch(RulePointer, St#yecc.rule_pointer2rule),
-    {Symbols, Line, N}.
+    #rule{n = N, anno = Anno, symbols = Symbols} =
+        maps:get(RulePointer, St#yecc.rule_pointer2rule),
+    {Symbols, Anno, N}.
 
 get_rule(RuleNmbr, St) ->
-    dict:fetch(RuleNmbr, St#yecc.rule_pointer2rule).
+    maps:get(RuleNmbr, St#yecc.rule_pointer2rule).
 
 tokens(RuleNmbr, St) ->
-    Rule = dict:fetch(RuleNmbr, St#yecc.rule_pointer2rule),
+    Rule = maps:get(RuleNmbr, St#yecc.rule_pointer2rule),
     Rule#rule.tokens.
 
 goto(From, Symbol, St) ->
@@ -2463,7 +2476,7 @@ include(St, File, Outport) ->
 include1(eof, _, _, _File, L, _St) ->
     L;
 include1({error, _}=_Error, _Inport, _Outport, File, L, St) ->
-    throw(add_error(File, L, cannot_parse, St));
+    throw(add_error(File, erl_anno:new(L), cannot_parse, St));
 include1(Line, Inport, Outport, File, L, St) ->
     Incr = case member($\n, Line) of
                true -> 1;
@@ -2488,7 +2501,7 @@ includefile_version(Includefile) ->
 
 parse_file(Epp) ->
     case epp:parse_erl_form(Epp) of
-        {ok, {function,_Line,yeccpars1,7,_Clauses}} ->
+        {ok, {function,_Anno,yeccpars1,7,_Clauses}} ->
             {1,4};
         {eof,_Line} ->
             {1,1};
@@ -2503,7 +2516,7 @@ pp_tokens(Tokens, Line0, Enc) ->
 pp_tokens1([], _Line0, _Enc, _T0) ->
     [];
 pp_tokens1([T | Ts], Line0, Enc, T0) ->
-    Line = element(2, T),
+    Line = location(anno(T)),
     [pp_sep(Line, Line0, T0), pp_symbol(T, Enc)|pp_tokens1(Ts, Line, Enc, T)].
 
 pp_symbol({var,_,Var}, _Enc) -> Var;
@@ -2538,10 +2551,17 @@ output_file_directive(St, _Filename, _Line) ->
     St.
 
 first_line(Tokens) ->
-    element(2, hd(Tokens)).
+    location(anno(hd(Tokens))).
 
 last_line(Tokens) ->
-    element(2, lists:last(Tokens)).
+    location(anno(lists:last(Tokens))).
+
+location(none) -> none;
+location(Anno) ->
+    erl_anno:line(Anno).
+
+anno(Token) ->
+    element(2, Token).
 
 %% Keep track of the current line in the generated file.
 fwrite(#yecc{outport = Outport, line = Line}=St, Format, Args) ->
@@ -2581,18 +2601,18 @@ format_symbol(Symbol) ->
     String = concat([Symbol]),
     case erl_scan:string(String) of
         {ok, [{atom, _, _}], _} ->
-            io_lib:fwrite(<<"~w">>, [Symbol]);
+            io_lib:fwrite(<<"~tw">>, [Symbol]);
         {ok, [{Word, _}], _} when Word =/= ':', Word =/= '->' ->
             case erl_scan:reserved_word(Word) of
                 true ->
                     String;
                 false ->
-                    io_lib:fwrite(<<"~w">>, [Symbol])
+                    io_lib:fwrite(<<"~tw">>, [Symbol])
             end;
         {ok, [{var, _, _}], _} ->
             String;
         _ -> 
-            io_lib:fwrite(<<"~w">>, [Symbol])
+            io_lib:fwrite(<<"~tw">>, [Symbol])
     end.
 
 inverse(L) ->
